import contextlib
import dataclasses
import hashlib
import itertools
import os
import subprocess
from typing import Iterable, List, Optional, Set, Any, Generator, Dict, Sequence, Tuple, Iterator, Callable

from debian.substvars import Substvars

from debputy.debhelper_emulation import dhe_install_pkg_file_as_ctrl_file_if_present, dhe_dbgsym_root_dir, \
    dhe_read_dbgsym_build_ids, dhe_read_dbgsym_migration
from debputy.filesystem_scan import FSPath, build_fs_from_root_dir
from debputy.highlevel_manifest import HighLevelManifest, PackageTransformationDefinition
from debputy.installdeb_emulation import dh_installdeb_emulation
from debputy.maintscript_snippet import UDEB_CONTROL_SCRIPTS, ALL_CONTROL_SCRIPTS
from debputy.packages import BinaryPackage
from debputy.packaging.makeshlibs import compute_shlibs, SONAMEInfo, extract_soname_info
from debputy.plugin.api import DpkgTriggerType
from debputy.plugin.api.impl import PluginProvidedTrigger, BinaryCtrlAccessorProvider, \
    MetadataDetectorPackageContextProvider, MetadataOrMaintscriptDetector
from debputy.plugin.api.spec import FlushableSubstvars
from debputy.util import _error, ensure_dir, assume_not_none

# For unclean builds, we need to unlink these to generate reproducible output.
_UNCLEAN_BUILD_UNLINK_CONTROL_FILES = [
    'control',
    'md5sums',
    'conffiles',
]


@dataclasses.dataclass(slots=True, frozen=True)
class BinaryPackageData:
    binary_package: BinaryPackage
    binary_staging_root_dir: str
    fs_root: Optional[FSPath]
    substvars: FlushableSubstvars
    shared_libraries: Optional[List[SONAMEInfo]]


def generate_md5sums_file(root_dir: str, fs_root: FSPath) -> None:
    conffiles = os.path.join(root_dir, 'DEBIAN', 'conffiles')
    md5sums = os.path.join(root_dir, 'DEBIAN', 'md5sums')
    exclude = set()
    if os.path.isfile(conffiles):
        with open(conffiles, 'rt') as fd:
            for line in fd:
                if not line.startswith("/"):
                    continue
                exclude.add('.' + line.rstrip("\n"))
    had_content = False
    files = (path for path in fs_root.all_paths() if path.is_file and path.path not in exclude)
    with open(md5sums, 'wt') as md5fd:
        for member in files:
            path = member.path
            assert path.startswith('./')
            path = path[2:]
            with open(member.fs_path, 'rb') as f:
                file_hash = hashlib.md5()
                while chunk := f.read(8192):
                    file_hash.update(chunk)
            had_content = True
            md5fd.write(f"{file_hash.hexdigest()}  {path}\n")
    if not had_content:
        os.unlink(md5sums)


def install_or_generate_conffiles(binary_package: BinaryPackage, root_dir: str, fs_root: FSPath) -> None:
    conffiles_dest = os.path.join(root_dir, 'DEBIAN', 'conffiles')
    dhe_install_pkg_file_as_ctrl_file_if_present(binary_package, 'conffiles', root_dir, 0o0644)
    etc_dir = fs_root.lookup('etc')
    if etc_dir:
        _add_conffiles(conffiles_dest, (p for p in etc_dir.all_paths() if p.is_file))
    if os.path.isfile(conffiles_dest):
        os.chmod(conffiles_dest, 0o0644)


def collect_data_from_binary_staging_dir(manifest: HighLevelManifest) -> Dict[str, BinaryPackageData]:
    package_data_table = {}

    for dctrl_bin in manifest.all_packages:
        package = dctrl_bin.name
        root_dir = os.path.join('debian', package)
        fs_root = None
        sonames = None
        if dctrl_bin.should_be_acted_on:
            fs_root = build_fs_from_root_dir(root_dir)
            manifest.apply_fs_transformations(package, fs_root)
            sonames = extract_soname_info(dctrl_bin, fs_root)
            substvars_file = f'debian/{package}.substvars'
            substvars = FlushableSubstvars.load_from_path(substvars_file, missing_ok=True)
            # We do not want to touch the substvars file (non-clean rebuild contamination)
            substvars.substvars_path = None
        else:
            substvars = FlushableSubstvars()

        package_data_table[package] = BinaryPackageData(
            dctrl_bin,
            root_dir,
            fs_root,
            substvars,
            sonames,
        )

    return package_data_table


def build_dbgsym_package_if_relevant(binary_package: BinaryPackage,
                                     dpkg_deb_args: List[str],
                                     output_dir: str,
                                     ) -> None:
    if binary_package.is_udeb:
        # We never built udebs due to #797391.  If it is easy to do, we can fix it when we do the control file
        return
    if not os.path.isdir(output_dir):
        _error(f"Cannot produce a dbgsym package when output path is not a directory.")
        return
    dbgsym_root = dhe_dbgsym_root_dir(binary_package)
    if not os.path.isdir(dbgsym_root):
        return
    package = binary_package.name
    if not os.path.isfile(os.path.join(dbgsym_root, 'DEBIAN', 'control')):
        # Will not be a problem when we integrate generation of DEBIAN/control (#18)
        _error(f"There is a dbgsym directory from debhelper for {package}, but not a DEBIAN/control")

    fs_root = build_fs_from_root_dir(dbgsym_root)
    generate_md5sums_file(dbgsym_root, fs_root)

    dpkg_cmd = [
        'dpkg-deb',
        '--root-owner-group',
    ]
    dpkg_cmd.extend(dpkg_deb_args)
    dpkg_cmd.extend([
        '--build',
        dbgsym_root,
        output_dir,
    ])

    print(f"Executing for {package}-dbgsym: {dpkg_cmd}")
    subprocess.check_call(dpkg_cmd)


def _filter_compress_paths(dctrl: BinaryPackage,
                           target_doc_package: BinaryPackage,
                           ) -> Callable[[FSPath], Iterator[FSPath]]:

    ignore_paths = {
        f'./usr/share/doc/{dctrl.name}/examples',
        f'./usr/share/doc/{target_doc_package.name}/examples',
    }
    ignore_dir_basenames = {
        '_sources',
    }
    ignore_basenames = {
        '.htaccess',
        'index.sgml',
        'objects.inv',
        'search_index.json',
        'copyright'
    }
    ignore_extensions = {
        '.xhtml',
        '.gif',
        '.png',
        '.jpg',
        '.jpeg',
        '.gz',
        '.taz',
        '.tgz',
        '.z',
        '.bz2',
        '.epub',
        '.jar',
        '.zip',
        '.odg',
        '.odp',
        '.odt',
        '.css',
        '.xz',
        '.lz',
        '.lzma',
        '.haddock',
        '.hs',
        '.woff',
        '.woff2',
        '.svg',
        '.svgz',
        '.js',
        '.devhelp2'
        '.map',  # Technically, dh_compress has this one case-sensitive
    }
    ignore_special_cases = (
        '-gz', '-z', '_z'
    )

    def _filtered_walk(path: FSPath) -> Iterator[FSPath]:
        for path, children in path.walk():
            if path.path in ignore_paths or path.name in ignore_dir_basenames:
                children.clear()
                continue
            name = path.name
            if path.is_symlink or not path.is_file or name in ignore_basenames or not path.has_fs_path:
                continue

            name_lc = name.lower()
            _, ext = os.path.splitext(name_lc)

            if ext in ignore_extensions or name_lc.endswith(ignore_special_cases) or path.stat().st_size <= 4096:
                continue
            yield path

    return _filtered_walk


def find_compressable_paths(dctrl: BinaryPackage,
                            target_doc_package: BinaryPackage,
                            fs_root: FSPath,
                            ) -> Iterator[FSPath]:

    path_filter = _filter_compress_paths(dctrl, target_doc_package)

    for p in ('./usr/share/info', './usr/share/man', './usr/share/doc'):
        path = fs_root.lookup(p)
        if path is None:
            continue
        yield from path_filter(path)
    x11_path = fs_root.lookup('./usr/share/fonts/X11')
    if x11_path:
        for path in x11_path.all_paths():
            if path.is_file and path.name.endswith('.pcf'):
                yield path


def usr_local_transformation(dctrl: BinaryPackage, fs_root: FSPath) -> None:
    path = fs_root.lookup('./usr/local')
    if path and any(path.iterdir):
        # There are two key issues:
        #  1) Getting the generated maintscript carried on to the final maintscript
        #  2) Making sure that manifest created directories do not trigger the "unused error".
        _error(f'Replacement of /usr/local paths is currently not supported in debputy (triggered by: {dctrl.name}).')


def apply_compression(dctrl: BinaryPackage, target_doc_package: BinaryPackage, fs_root: FSPath) -> None:
    # TODO: Support hardlinks
    compressed_files: Dict[str, str] = {}
    for path in find_compressable_paths(dctrl, target_doc_package, fs_root):
        parent_dir = path.parent_dir
        with parent_dir.add_file(f'{path.name}.gz') as new_file, open(new_file.fs_path, 'wb') as fd:
            subprocess.check_call(['gzip', '-9nc', path.fs_path], stdout=fd)
            compressed_files[path.path] = new_file.path
        del parent_dir[path.name]

    all_remaining_symlinks = {p.path: p for p in fs_root.all_paths() if p.is_symlink}
    changed = True
    while changed:
        changed = False
        remaining: List[FSPath] = list(all_remaining_symlinks.values())
        for symlink in remaining:
            target = symlink.readlink()
            dir_target, basename_target = os.path.split(target)
            new_basename_target = f'{basename_target}.gz'
            dir_path = fs_root.lookup(dir_target)
            if not dir_path or basename_target in dir_path or new_basename_target not in dir_path:
                continue
            del all_remaining_symlinks[symlink.path]
            changed = True

            new_link_name = f'{symlink.name}.gz' if not symlink.name.endswith('.gz') else symlink.name
            symlink.parent_dir.add_symlink(new_link_name, os.path.join(dir_target, new_basename_target))
            symlink.parent_dir = None


def setup_control_files(binary_package_data: BinaryPackageData,
                        manifest: HighLevelManifest,
                        related_udeb_package_data: Optional[BinaryPackageData],
                        ) -> None:
    binary_package = assume_not_none(binary_package_data.binary_package)
    root_dir = binary_package_data.binary_staging_root_dir
    fs_root = assume_not_none(binary_package_data.fs_root)
    soname_info_list = assume_not_none(binary_package_data.shared_libraries)
    ensure_dir(os.path.join(root_dir, 'DEBIAN'))
    package_state = manifest.package_state_for(binary_package.name)

    dbgsym_root_dir = dhe_dbgsym_root_dir(binary_package)

    for ctrl_file_to_be_removed in _UNCLEAN_BUILD_UNLINK_CONTROL_FILES:
        try:
            os.unlink(os.path.join(root_dir, 'DEBIAN', ctrl_file_to_be_removed))
        except FileNotFoundError:
            pass
        try:
            os.unlink(os.path.join(dbgsym_root_dir, 'DEBIAN', ctrl_file_to_be_removed))
        except FileNotFoundError:
            pass

    generated_triggers = set()
    plugin_provided_triggers: Dict[Tuple[DpkgTriggerType, str], PluginProvidedTrigger] = {}
    metadata_maintscript_detectors = manifest.plugin_provided_feature_set.metadata_maintscript_detectors
    package_metadata_context = MetadataDetectorPackageContextProvider(
        binary_package,
        related_udeb_package_data.binary_package if related_udeb_package_data else None,
        # FIXME: source_package
    )

    if not binary_package.is_arch_all:
        udeb_package_name = related_udeb_package_data.binary_package.name if related_udeb_package_data else None
        compute_shlibs(binary_package,
                       root_dir,
                       fs_root,
                       manifest,
                       soname_info_list,
                       udeb_package_name,
                       generated_triggers,
                       )

    state = manifest.package_state_for(binary_package.name)
    substvars = binary_package_data.substvars

    maintscript_snippets = {k: v.copy() for k, v in state.maintscript_snippets.items()}
    plugin_detector_definition: MetadataOrMaintscriptDetector
    for plugin_detector_definition in itertools.chain.from_iterable(metadata_maintscript_detectors.values()):
        if not plugin_detector_definition.applies_to(binary_package):
            continue
        ctrl = BinaryCtrlAccessorProvider(
            plugin_detector_definition.plugin_metadata,
            package_metadata_context,
            plugin_detector_definition.detector_id,
            plugin_provided_triggers,
            substvars,
            maintscript_snippets,
            state.substitution,
        )
        plugin_detector_definition.run_detector(fs_root, ctrl, package_metadata_context)

    if binary_package.is_udeb:
        _generate_control_files(binary_package, package_state, root_dir, fs_root, substvars)
        for script in UDEB_CONTROL_SCRIPTS:
            dhe_install_pkg_file_as_ctrl_file_if_present(binary_package, script, root_dir, 0o0755)
        return

    dh_installdeb_emulation(binary_package, root_dir, generated_triggers, maintscript_snippets)
    install_or_generate_conffiles(binary_package, root_dir, fs_root)
    _generate_control_files(binary_package, package_state, root_dir, fs_root, substvars)
    generate_md5sums_file(root_dir, fs_root)


def _add_conffiles(conffiles_dest: str, conffile_matches: Iterable[FSPath]) -> None:
    with open(conffiles_dest, 'at') as fd:
        for conffile_match in conffile_matches:
            conffile = conffile_match.path
            assert conffile_match.is_file
            assert conffile.startswith('./')
            fd.write(f"{conffile[1:]}\n")
    if os.stat(conffiles_dest).st_size == 0:
        os.utime(conffiles_dest)


@contextlib.contextmanager
def _ensure_base_substvars_defined(substvars: Substvars) -> None:
    for substvar in ('misc:Depends', 'misc:Pre-Depends'):
        if substvar not in substvars:
            substvars[substvar] = ''


def _compute_installed_size(fs_root: FSPath) -> int:
    """Emulate dpkg-gencontrol's code for computing the default Installed-Size"""
    size_in_kb = 0
    hard_links = set()
    for path in fs_root.all_paths():
        if not path.is_dir and path.has_fs_path:
            st = path.stat()
            if st.st_nlink > 1:
                hl_key = (st.st_dev, st.st_ino)
                if hl_key in hard_links:
                    continue
                hard_links.add(hl_key)
            path_size = (st.st_size + 1023) // 1024
        elif path.is_symlink:
            path_size = (len(path.readlink()) + 1023) // 1024
        else:
            path_size = 1
        size_in_kb += path_size
    return size_in_kb


def _generate_dbgsym_control_file_if_relevant(binary_package: BinaryPackage,
                                              dbgsym_root_dir: str,
                                              substvars_file: str,
                                              dbgsym_ids: str,
                                              multi_arch: Optional[str],
                                              extra_common_params: Sequence[str],
                                              ) -> None:
    section = binary_package.archive_section
    component = ''
    extra_params = []
    migration = dhe_read_dbgsym_migration(binary_package)
    if section is not None and '/' in section and not section.startswith('main/'):
        component = section.split('/', 1)[1] + '/'
    if multi_arch != 'same':
        extra_params.append('-UMulti-Arch')
    if migration:
        extra_params.append(f"-DReplaces={migration}")
        extra_params.append(f"-DBreaks={migration}")
    else:
        extra_params.append("-UReplaces")
        extra_params.append("-UBreaks")
    extra_params.extend(extra_common_params)

    ensure_dir(os.path.join(dbgsym_root_dir, 'DEBIAN'))
    package = binary_package.name
    dpkg_cmd = [
        'dpkg-gencontrol',
        f'-p{package}',
        '-ldebian/changelog',
        f'-T{substvars_file}',
        f'-P{dbgsym_root_dir}',
        f'-DPackage={package}-dbgsym',
        '-DDepends=' + package + ' (= ${binary:Version})',
        f'-DDescription=debug symbols for {package}',
        f'-DSection={component}debug',
        f'-DBuild-Ids={dbgsym_ids}',
        '-UPre-Depends',
        '-URecommends',
        '-USuggests',
        '-UEnhances',
        '-UProvides',
        '-UEssential',
        '-UConflicts',
        '-DPriority=optional',
        '-UHomepage',
        '-UImportant',
        '-UBuilt-Using',
        '-DAuto-Built-Package=debug-symbols',
        '-UProtected',
        *extra_params,

    ]
    print(f'Executing for {package}-dbgsym: {dpkg_cmd}')
    subprocess.check_call(dpkg_cmd)
    os.chmod(os.path.join(dbgsym_root_dir, 'DEBIAN', 'control'), 0o644)


def _all_parent_directories_of(directories: Iterable[str]) -> Set[str]:
    result = {'.'}
    for path in directories:
        current = os.path.dirname(path)
        while current and current not in result:
            result.add(current)
            current = os.path.dirname(current)
    return result


def _auto_compute_multi_arch(binary_package: BinaryPackage,
                             package_root_dir: str,
                             fs_root: FSPath,
                             ) -> Optional[str]:
    resolved_arch = binary_package.resolved_architecture
    if resolved_arch == 'all':
        return None
    ctrl_dir = os.path.join(package_root_dir, 'DEBIAN')
    if any(script for script in ALL_CONTROL_SCRIPTS if os.path.isfile(os.path.join(ctrl_dir, script))):
        return None

    resolved_multiarch = binary_package.deb_multiarch
    assert resolved_arch != 'all'
    acceptable_no_descend_paths = {
        f'./usr/lib/{resolved_multiarch}',
        f'./usr/include/{resolved_multiarch}',
    }
    acceptable_files = {
        f'./usr/share/doc/{binary_package.name}/{basename}'
        for basename in (
            'copyright',
            'changelog.gz',
            'changelog.Debian.gz',
            f'changelog.Debian.{resolved_arch}.gz'
            'NEWS.Debian',
            'NEWS.Debian.gz',
            'README.Debian',
            'README.Debian.gz',
        )
    }
    acceptable_intermediate_dirs = _all_parent_directories_of(itertools.chain(acceptable_no_descend_paths,
                                                                              acceptable_files))

    for fs_path, children in fs_root.walk():
        path = fs_path.path
        if path in acceptable_no_descend_paths:
            children.clear()
            continue
        if path in acceptable_intermediate_dirs or path in acceptable_files:
            continue
        return None

    return 'same'


def _generate_control_files(binary_package: BinaryPackage,
                            package_state: PackageTransformationDefinition,
                            package_root_dir: str,
                            fs_root: FSPath,
                            substvars: FlushableSubstvars,
                            ) -> None:
    package = binary_package.name
    extra_common_params = []
    extra_params_specific = []
    _ensure_base_substvars_defined(substvars)
    if 'Installed-Size' not in substvars:
        # Pass it via cmd-line to make it more visible that we are providing the
        # value.  It also prevents the dbgsym package from picking up this value.
        ctrl_fs_root = build_fs_from_root_dir(os.path.join(package_root_dir, 'DEBIAN'))
        total_size = _compute_installed_size(fs_root) + _compute_installed_size(ctrl_fs_root)
        extra_params_specific.append(f'-VInstalled-Size={total_size}')

    ma_value = binary_package.fields.get('Multi-Arch')
    if not binary_package.is_udeb and ma_value is None:
        ma_value = _auto_compute_multi_arch(binary_package, package_root_dir, fs_root)
        if ma_value is not None:
            print(f'The package "{binary_package.name}" looks like it should be "Multi-Arch: {ma_value}" based'
                  ' on the contents and there is no explicit "Multi-Arch" field. Setting the Multi-Arch field'
                  ' accordingly in the binary.  If this auto-correction is wrong, please  add "Multi-Arch: no" to the'
                  ' relevant part of "debian/control" to disable this feature.')
            extra_params_specific.append(f'-DMulti-Arch={ma_value}')
    elif ma_value == 'no':
        extra_params_specific.append(f'-UMulti-Arch')

    dbgsym_root_dir = dhe_dbgsym_root_dir(binary_package)
    dbgsym_ids = dhe_read_dbgsym_build_ids(binary_package)
    if package_state.binary_version is not None:
        extra_common_params.append(f'-v{package_state.binary_version}')

    with substvars.flush() as flushed_substvars:
        if os.path.isdir(dbgsym_root_dir):
            _generate_dbgsym_control_file_if_relevant(binary_package,
                                                      dbgsym_root_dir,
                                                      flushed_substvars,
                                                      dbgsym_ids,
                                                      ma_value,
                                                      extra_common_params,
                                                      )
        elif dbgsym_ids:
            extra_common_params.append(f'-DBuild-Ids={dbgsym_ids}')
        dpkg_cmd = [
            'dpkg-gencontrol',
            f'-p{package}',
            '-ldebian/changelog',
            f'-T{flushed_substvars}',
            f'-P{package_root_dir}',
            *extra_common_params,
            *extra_params_specific,
        ]
        print(f'Executing for {package}: {dpkg_cmd}')
        subprocess.check_call(dpkg_cmd)
        os.chmod(os.path.join(package_root_dir, 'DEBIAN', 'control'), 0o644)
