import contextlib
import dataclasses
import errno
import os.path
import re
import shutil
from re import Match
from typing import Dict, Optional, Callable, Union, Iterable, TextIO, Tuple, Generator, Any, Literal, Sequence, List

from debputy.maintscript_snippet import STD_CONTROL_SCRIPTS, MaintscriptSnippet, MaintscriptSnippetContainer
from debputy.packages import BinaryPackage
from debputy.substitution import Substitution
from debputy.util import _escape_shell_word, ensure_dir, debputy_autoscript_tagline

SnippetReplacement = Union[str, Callable[[str], str]]
MAINTSCRIPT_TOKEN_NAME_PATTERN = r'[A-Za-z0-9_.+]+'
MAINTSCRIPT_TOKEN_NAME_REGEX = re.compile(MAINTSCRIPT_TOKEN_NAME_PATTERN)
MAINTSCRIPT_TOKEN_REGEX = re.compile(f'#({MAINTSCRIPT_TOKEN_NAME_PATTERN})#')
_ARCH_FILTER_START = re.compile(r'^(\[([^]]*)])[ \t]+')
_ARCH_FILTER_END = re.compile(r'\s+(\[([^]]*)])$')
_BUILD_PROFILE_FILTER = re.compile(r'(<([^>]*)>(?:\s+<([^>]*)>)*)')


class CannotEmulateExecutableDHConfigFile(Exception):

    def message(self) -> str:
        return self.args[0]

    def config_file(self) -> str:
        return self.args[1]


@dataclasses.dataclass(slots=True, frozen=True)
class DHConfigFileLine:
    config_file: str
    line_no: int
    executable_config: bool
    original_line: str
    tokens: Sequence[str]
    arch_filter: Optional[str]
    build_profile_filter: Optional[str]


@dataclasses.dataclass(frozen=True, slots=True)
class MaintscriptContent:
    content: str
    is_file: bool


def dhe_generated_file(binary_package: BinaryPackage, filename: str, mkdirs: bool = True) -> str:
    parent_dir = os.path.join('debian/.debhelper/generated/', binary_package.name)
    if mkdirs:
        ensure_dir(parent_dir)
    return os.path.join(parent_dir, filename)


def dhe_dbgsym_root_dir(binary_package: BinaryPackage) -> str:
    return os.path.join('debian', '.debhelper', binary_package.name, 'dbgsym-root')


def _concat_slurp_files(*files: MaintscriptContent) -> str:
    result = ''
    for content in files:
        if content.is_file:
            with open(content.content, 'rt') as fd:
                result += fd.read()
        else:
            result += content.content
        if not result.endswith('\n'):
            result += "\n"
    return result


@contextlib.contextmanager
def _lines_from(config_file: str) -> Generator[Tuple[TextIO, bool], Any, Any]:
    with open(config_file, 'rt') as fd:
        yield fd, os.access(config_file, os.X_OK)


def _prune_match(line: str,
                 match: Optional[Match[str]],
                 match_mapper: Optional[Callable[[Match[str]], str]] = None,
                 ) -> Tuple[str, Optional[str]]:
    if match is None:
        return line, None
    s, e = match.span()
    if match_mapper:
        matched_part = match_mapper(match)
    else:
        matched_part = line[s:e]
    # We prune exactly the matched part and assume the regexes leaves behind spaces if they were important.
    line = line[:s] + line[e:]
    # One special-case, if the match is at the beginning or end, then we can safely discard left
    # over whitespace.
    return line.strip(), matched_part


def dhe_filedoublearray(config_file: str, substitution: Optional[Substitution]) -> Iterable[DHConfigFileLine]:
    with _lines_from(config_file) as (fd, is_executable):
        for line_no, orig_line in enumerate(fd, start=1):
            arch_filter = None
            build_profile_filter = None
            if (line_no == 1 and is_executable and
                    not orig_line.startswith(("#!/usr/bin/dh-exec", "#! /usr/bin/dh-exec"))):
                raise CannotEmulateExecutableDHConfigFile("The #!-line must be #!/usr/bin/dh-exec", config_file)
            orig_line = orig_line.rstrip("\n")
            line = orig_line.strip()
            if not line or line.startswith("#"):
                continue
            if is_executable:
                if '=>' in line:
                    raise CannotEmulateExecutableDHConfigFile('Cannot emulate dh-exec\'s "=>" feature to rename files',
                                                              config_file
                                                              )
                line, arch_filter = _prune_match(line,
                                                 _ARCH_FILTER_START.match(line) or _ARCH_FILTER_END.match(line),
                                                 # Remove the enclosing []
                                                 lambda m: m.group(1)[1:-1].strip(),
                                                 )
                line, build_profile_filter = _prune_match(line,
                                                          _BUILD_PROFILE_FILTER.match(line),
                                                          )
            parts = tuple(substitution.substitute(w, f'{config_file} line {line_no} token "{w}"') for w in line.split())
            yield DHConfigFileLine(config_file,
                                   line_no,
                                   is_executable,
                                   orig_line,
                                   parts,
                                   arch_filter,
                                   build_profile_filter,
                                   )


def dhe_pkgfile(binary_package: BinaryPackage,
                basename: str,
                always_fallback_to_packageless_variant: bool = False,
                mandatory: bool = False
                ) -> Optional[str]:
    # TODO: Architecture specific files
    possible_names = [
        "%s.%s" % (binary_package.name, basename)
    ]
    if binary_package.is_main_package or always_fallback_to_packageless_variant:
        possible_names.append(basename)

    for name in possible_names:
        path = "debian/%s" % name
        if os.path.exists(path):
            return path
    if mandatory:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), os.path.join('debian', possible_names[0]))
    return None


def dhe_install_pkg_file_as_ctrl_file_if_present(binary_package: BinaryPackage,
                                                 basename: str,
                                                 root_dir: str,
                                                 mode: int,
                                                 **kwargs,
                                                 ) -> None:
    source = dhe_pkgfile(binary_package, basename,  **kwargs)
    if source is None:
        return
    ctrl_dir = os.path.join(root_dir, 'DEBIAN')
    ensure_dir(ctrl_dir)
    dhe_install_path(source, os.path.join(ctrl_dir, basename), mode)


def dhe_install_path(source: str, dest: str, mode: int) -> None:
    # TODO: "install -p -mXXXX foo bar" silently discards broken
    # symlinks to install the file in place.  (#868204)
    print(f"    install -p -m{oct(mode)[2:]} {_escape_shell_word(source)} {_escape_shell_word(dest)}")
    shutil.copyfile(source, dest)
    os.chmod(dest, mode)


def _read_dbgsym_file(binary_package: BinaryPackage, dbgsym_file: str) -> Optional[str]:
    p = os.path.join('debian', '.debhelper', binary_package.name, dbgsym_file)
    if os.path.isfile(p):
        with open(p, 'rt') as fd:
            return fd.read().strip()
    return None


def dhe_read_dbgsym_migration(binary_package: BinaryPackage) -> Optional[str]:
    return _read_dbgsym_file(binary_package, 'dbgsym-migration')


def dhe_read_dbgsym_build_ids(binary_package: BinaryPackage) -> str:
    raw = _read_dbgsym_file(binary_package, 'dbgsym-build-ids')
    if raw is None:
        return ''
    seen = set()

    def _unique_build_ids() -> Iterable[str]:
        for bid in sorted(raw.split()):
            if bid in seen:
                continue
            seen.add(bid)
            yield bid

    return " ".join(_unique_build_ids())


def dhe_autoscript_expanded_snippet(binary_package: BinaryPackage,
                                    script: str,
                                    snippet: str,
                                    tool_with_version: str,
                                    snippet_order: Optional[Literal['service']] = None
                                    ) -> None:

    assert snippet_order is None or snippet_order == 'service'
    assert script in STD_CONTROL_SCRIPTS
    if snippet_order is None:
        outfile = f'debian/{binary_package.name}.{script}.debhelper'
    else:
        outfile = dhe_generated_file(binary_package, f"{script}.{snippet_order}")
    reverse_order = script in ('postrm', 'prerm')
    if not snippet.endswith("\n"):
        snippet += "\n"
    combined_snippet = (
            f"# Automatically added by {tool_with_version}\n"
            + snippet
            + "# End automatically added section\n"
    )
    if reverse_order:
        with open(outfile, 'wt') as wfd, open(outfile, 'rt') as rfd:
            wfd.write(combined_snippet)
            shutil.copyfileobj(rfd, wfd)
    else:
        with open(outfile, 'at') as fd:
            fd.write(combined_snippet)


def _make_script_replacement_func(variables: Dict[str, SnippetReplacement],
                                  ) -> Callable[[re.Match[str]], str]:
    cache: Dict[str, str] = {}

    def _impl(m: re.Match[str]) -> str:
        variable = m.group(1)
        if variable in cache:
            return cache[variable]
        if variable in variables:
            result = variables[variable]
            if isinstance(result, Callable):
                result = result(variable)
            cache[variable] = result
            return result
        return f'#{variable}#'

    return _impl


def _generate_script_subst_table(package: str,
                                 debhelper_snippet_files: Iterable[MaintscriptContent],
                                 extra_variables: Optional[Dict[str, SnippetReplacement]],
                                 ) -> Dict[str, SnippetReplacement]:
    variables: Dict[str, SnippetReplacement] = {}
    if extra_variables:
        variables.update(extra_variables)
        # "pkg.<pkg>." is 5 + the length of <pkg>
        subst_len = 5 + len(package)
        # Match debhelper's "pkg.<pkg>.FOO" => "FOO" rule
        for k, v in extra_variables.items():
            if not MAINTSCRIPT_TOKEN_NAME_REGEX.match(k):
                raise ValueError(f'User defined token "{k}" does not match {MAINTSCRIPT_TOKEN_NAME_PATTERN}')
            if k.startswith(f'pkg.{package}.') and len(k) > subst_len:
                variables[k[subst_len:]] = v
    if 'PACKAGE' not in variables:
        variables['PACKAGE'] = package
    if 'DEBHELPER' not in variables:
        variables['DEBHELPER'] = lambda _: _concat_slurp_files(*debhelper_snippet_files)
    return variables


def debhelper_script_subst(binary_package: BinaryPackage,
                           root_dir: str,
                           script: str,
                           maintscript_snippets: Dict[str, MaintscriptSnippetContainer],
                           extra_variables: Optional[Dict[str, SnippetReplacement]],
                           ) -> None:
    assert script in STD_CONTROL_SCRIPTS
    package = binary_package.name
    user_provided_file = dhe_pkgfile(binary_package, script)
    ensure_dir(os.path.join(root_dir, 'DEBIAN'))
    dest = os.path.join(root_dir, 'DEBIAN', script)
    debputy_snippets = maintscript_snippets.get(script)
    snippets = []
    tag_line = debputy_autoscript_tagline()
    dh_generic_snippets = f'debian/{package}.{script}.debhelper'
    dh_service_snippets = dhe_generated_file(binary_package, f"{script}.service", mkdirs=False)

    if os.path.isfile(dh_generic_snippets):
        snippets.append(MaintscriptContent(content=dh_generic_snippets, is_file=True))

    if debputy_snippets and debputy_snippets.has_content():
        snippets.append(MaintscriptContent(content=debputy_snippets.generate_snippet(tag_line), is_file=False))

    if os.path.isfile(dh_service_snippets):
        snippets.append(MaintscriptContent(content=dh_service_snippets, is_file=True))

    if debputy_snippets and debputy_snippets.has_content(snippet_order='service'):
        snippets.append(MaintscriptContent(content=debputy_snippets.generate_snippet(tag_line, snippet_order='service'),
                                           is_file=False))

    has_snippets = bool(snippets)
    if script in ('prerm', 'postrm'):
        snippets = reversed(snippets)

    variables = _generate_script_subst_table(package, snippets, extra_variables)
    if user_provided_file:
        with open(dest, 'wt') as dest_fd, open(user_provided_file, 'rt') as source_fd:
            replacement = _make_script_replacement_func(variables)
            for line in source_fd:
                dest_fd.write(MAINTSCRIPT_TOKEN_REGEX.sub(replacement, line))
    elif has_snippets:
        with open(dest, 'wt') as fd:
            fd.write("#!/bin/sh\n")
            fd.write("set -e\n")
            fd.write(_concat_slurp_files(*snippets))
    else:
        return
    os.chmod(dest, 0o755)
