import json
import os
import subprocess
from typing import Optional, List

from debputy.dh_migration.migrators import MIGRATORS
from debputy.dh_migration.models import FeatureMigration, AcceptableMigrationIssues, UnsupportedFeature, \
    ConflictingChange
from debputy.highlevel_manifest import HighLevelManifest
from debputy.util import _error


def _print_migration_summary(migrations: List[FeatureMigration],
                             permit_destructive_changes: bool,
                             compat: int,
                             min_compat_level: int,
                             ) -> None:
    warning_count = 0
    print("MIGRATION SUMMERY")
    print("=================")
    print()

    for migration in migrations:
        print()
        if not migration.anything_to_do:
            print(f'Migration: {migration.tagline} - Nothing to do and no issues detected!')
            continue
        underline = '-' * len(migration.tagline)
        print(f'Summary for migration: {migration.tagline}')
        print(f'-----------------------{underline}')
        print()
        if migration.warnings:
            print(" /!\\ ATTENTION /!\\")
            warning_count += len(migration.warnings)
            for warning in migration.warnings:
                print(f"    * {warning}")
        else:
            print(f'  No warnings detected / manual conversions needed')
        print()
        if permit_destructive_changes:
            print(f' * Manifest changes applied: {migration.successful_changes}')
            print(f' * Settings already present: {migration.already_present}')
            print(f' * Files removed (merged): {migration.remove_paths_on_success}')
            print(f' * Files renamed: {[" -> ".join(p) for p in migration.rename_paths_on_success]}')
        else:
            print(f' * Manifest changes to be applied: {migration.successful_changes}')
            print(f' * Settings already present: {migration.already_present}')
            print(f' * Files to be removed (merged): {migration.remove_paths_on_success}')
            print(f' * Files to be renamed: {[" -> ".join(p) for p in migration.rename_paths_on_success]}')

    if compat < min_compat_level:
        warning_count += 1
        print()
        print(f"WARNING: The migration tool assumes debhelper compat {min_compat_level}+ semantics, but this package"
              f" is using compat {compat}.  Consider upgrading the package to compat {min_compat_level}"
              " first.")

    if warning_count:
        print()
        print(f'/!\\ Total number of warnings or manual migrations required: {warning_count}')


def _dh_compat_level() -> Optional[int]:
    try:
        res = subprocess.check_output(['dh_assistant', 'active-compat-level'], stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        compat = None
    else:
        try:
            compat = json.loads(res)["declared-compat-level"]
        except RuntimeError:
            compat = None
        else:
            if not isinstance(compat, int):
                compat = None
    return compat


def migrate_from_dh(manifest: HighLevelManifest,
                    acceptable_migration_issues: AcceptableMigrationIssues,
                    permit_destructive_changes: bool) -> None:
    migrations = []
    compat = _dh_compat_level()
    # Compat 14 due to dh_installpam installation directory
    # Compat 13 due to dh_installtmpfiles (see TODO in migrate_tmpfile)
    min_compat = 14
    if compat is None:
        _error(f'Cannot detect declared compat level (try running "dh_assistant active-compat-level")')
        return

    if compat < min_compat and 'min-compat-level' not in acceptable_migration_issues:
        # The migration summary special-cases the compat mismatch and warns for us.
        _error(f'The migration tool assumes debhelper compat {min_compat} or later but the package is only on'
               f' compat {compat}.  This may lead to incorrect result.'
               f"\n\nUse --acceptable-migration-issues=min-compat-level to convert this into a warning and"
               f" try again, if you want to continue regardless.")

    try:
        for migrator in MIGRATORS:
            migrations.extend(migrator(manifest, acceptable_migration_issues))
    except UnsupportedFeature as e:
        msg = f'Unable to migrate automatically due to missing features in debputy. The feature is:' \
              f'\n\n  * {e.message}'
        keys = e.issue_keys
        if keys:
            primary_key = keys[0]
            alt_keys = ''
            if len(keys) > 1:
                alt_keys = f' Alternatively you can also use one of: {", ".join(keys[1:])}.  Please note that some' \
                           ' of these may cover more cases.'
            msg += (f"\n\nUse --acceptable-migration-issues={primary_key} to convert this into a warning and try again."
                    " However, you should only do that if you believe you can replace the functionality manually"
                    f" or the usage is obsolete / can be removed. {alt_keys}"
                    )
        _error(msg)
    except ConflictingChange as e:
        _error("The migration tool detected a conflict data being migrated and data already migrated / in the existing"
               "manifest."
               f"\n\n   * {e.message}"
               '\n\nPlease review the situation and resolve the conflict manually.')

    _print_migration_summary(migrations, permit_destructive_changes, compat, min_compat)
    manifest_change_count = sum((m.successful_changes for m in migrations), 0)

    if not manifest_change_count:
        print()
        print(f"debputy was not able to find any (supported) migrations that it could perform for you.")
        return

    if any(m.migration_changes_manifest and m.successful_changes for m in migrations):
        with open(manifest.manifest_path + ".new", 'w') as fd:
            manifest.mutable_manifest.write_to(fd)

        if permit_destructive_changes:
            if os.path.isfile(manifest.manifest_path):
                os.rename(manifest.manifest_path, manifest.manifest_path + '.orig')
            os.rename(manifest.manifest_path + '.new', manifest.manifest_path)
            print()
            print(f"Updated manifest {manifest.manifest_path}")
        else:
            print()
            print(f'Created draft manifest "{manifest.manifest_path}.new" (rename to "{manifest.manifest_path}"'
                  ' to activate it)')
    else:
        print()
        print('No manifest changes detected; skipping update of manifest.')

    removals = sum((len(m.remove_paths_on_success) for m in migrations), 0)
    renames = sum((len(m.rename_paths_on_success) for m in migrations), 0)

    if permit_destructive_changes:
        for previous_path, new_path in (p for m in migrations for p in m.rename_paths_on_success):
            os.rename(previous_path, new_path)
        for path in (p for m in migrations for p in m.remove_paths_on_success):
            os.unlink(path)

        print()
        print(f"Removed {removals} and renamed {renames} file/files")
    else:
        print()
        print(f"(Kept {removals + renames} file/files in place as requested; removals: {removals}, renames: {renames})")
