import json
import os
import subprocess
from typing import Iterable, Optional, Tuple, List

from debputy.architecture_support import dpkg_architecture_table
from debputy.debhelper_emulation import dhe_filedoublearray, DHConfigFileLine, dhe_pkgfile
from debputy.dh_migration.models import ConflictingChange, FeatureMigration, UnsupportedFeature, \
    AcceptableMigrationIssues, DHMigrationSubstitution
from debputy.highlevel_manifest import MutableYAMLSymlink, HighLevelManifest, MutableYAMLConffileManagementItem, \
    MutableCondition
from debputy.installdeb_emulation import _validate_rm_mv_conffile
from debputy.packages import BinaryPackage
from debputy.util import _error


# Align with debputy.py
DH_COMMANDS_REPLACED = frozenset({
    'dh_installcron',
    'dh_installifupdown',
    'dh_installtmpfiles',
    'dh_installmime',
    'dh_installmodules',
    'dh_installlogcheck',
    'dh_installlogrotate',
    'dh_installpam',
    'dh_installppp',
    'dh_installudev',
    'dh_installgsettings',
    'dh_installinitramfs',
    'dh_bugfiles',
    'dh_lintian',
    'dh_icons',
    'dh_link',
    'dh_installxfonts',
    'dh_compress',
    'dh_fixperms',
    'dh_makeshlibs',
    'dh_shlibdeps',
    'dh_installdeb',
    'dh_gencontrol',
    'dh_md5sums',
    'dh_builddeb',
})


def _extract_condition_from_dh_config_line(dh_config_line: DHConfigFileLine,
                                           acceptable_migration_issues: AcceptableMigrationIssues,
                                           warnings: List[str]):
    if dh_config_line.build_profile_filter and dh_config_line.arch_filter:
        if 'dh-exec-multiple-filters' in acceptable_migration_issues:
            warnings.append('TODO: MANUAL MIGRATION of conditions from executable dh config'
                            f' "{dh_config_line.config_file}" line {dh_config_line.line_no} is required')
            return None
        raise UnsupportedFeature('The dh-exec emulation cannot support multiple filters at the moment'
                                 f' (found: {dh_config_line.config_file} at line {dh_config_line.line_no}',
                                 ['dh-exec-multiple-filters'],
                                 )
    if dh_config_line.build_profile_filter:
        return MutableCondition.build_profiles_matches(dh_config_line.build_profile_filter)
    if dh_config_line.arch_filter:
        return MutableCondition.arch_matches(dh_config_line.arch_filter)
    return None


def _dh_config_file(dctrl_bin: BinaryPackage,
                    basename,
                    helper_name,
                    acceptable_migration_issues: AcceptableMigrationIssues,
                    warnings: List[str],
                    paths_to_remove: Optional[List[str]],
                    support_executable_files=False,
                    ) -> Tuple[Optional[str], Optional[Iterable[DHConfigFileLine]]]:
    dh_config_file = dhe_pkgfile(dctrl_bin, basename)
    if dh_config_file is None:
        return None, None
    if os.access(dh_config_file, os.X_OK) and not support_executable_files:
        primary_key = f'executable-{helper_name}-config'
        if (primary_key in acceptable_migration_issues
                or 'any-executable-dh-configs' in acceptable_migration_issues):
            warnings.append(f'TODO: MANUAL MIGRATION of executable dh config "{dh_config_file}" is required.')
            return None, None
        raise UnsupportedFeature(f'Executable configuration files not supported (found: {dh_config_file}).',
                                 [primary_key, 'any-executable-dh-configs'],
                                 )

    if paths_to_remove is not None:
        paths_to_remove.append(dh_config_file)
    substitution = DHMigrationSubstitution(
        dpkg_architecture_table(),
        acceptable_migration_issues,
        warnings,
    )
    content = dhe_filedoublearray(dh_config_file, substitution)
    return dh_config_file, content


def migrate_maintscript(manifest: HighLevelManifest,
                        acceptable_migration_issues: AcceptableMigrationIssues,
                        ) -> Iterable[FeatureMigration]:
    mutable_manifest = manifest.mutable_manifest
    paths_to_remove = []
    warnings = []
    manifest_changes = 0
    already_present = 0
    for dctrl_bin in manifest.all_packages:
        mainscript_file, content = _dh_config_file(dctrl_bin,
                                                   'maintscript',
                                                   'dh_installdeb',
                                                   acceptable_migration_issues,
                                                   warnings,
                                                   paths_to_remove,
                                                   )

        if mainscript_file is None:
            continue
        assert content is not None

        package_definition = mutable_manifest.package(dctrl_bin.name)
        conffiles = {it.obsolete_conffile: it for it in package_definition.conffile_management_items()}
        seen_conffiles = set()

        for dhe_line in content:
            cmd = dhe_line.tokens[0]
            if cmd not in {'rm_conffile', 'mv_conffile'}:
                raise UnsupportedFeature(f'The dh_installdeb file {mainscript_file} contains the (currently)'
                                         f' unsupported command "{cmd}" on line {dhe_line.line_no}'
                                         f' (line: "{dhe_line.original_line}"')

            try:
                _, obsolete_conffile, new_conffile, prior_to_version, owning_package = \
                    _validate_rm_mv_conffile(dctrl_bin.name, dhe_line)
            except ValueError as e:
                _error(f"Validation error in {mainscript_file} on line {dhe_line.line_no}. The error was: {e.args[0]}.")

            if obsolete_conffile in seen_conffiles:
                raise ConflictingChange(f'The {mainscript_file} file defines actions for "{obsolete_conffile}" twice!'
                                        f' Please ensure that it is defined at most once in that file.')
            seen_conffiles.add(obsolete_conffile)

            if cmd == 'rm_conffile':
                item = MutableYAMLConffileManagementItem.rm_conffile(
                    obsolete_conffile,
                    prior_to_version,
                    owning_package,
                )
            else:
                assert cmd == 'mv_conffile'
                item = MutableYAMLConffileManagementItem.mv_conffile(
                    obsolete_conffile,
                    new_conffile,
                    prior_to_version,
                    owning_package,
                )

            existing_def = conffiles.get(item.obsolete_conffile)
            if existing_def is not None:
                if not (
                        item.command == existing_def.command
                        and item.new_conffile == existing_def.new_conffile
                        and item.prior_to_version == existing_def.prior_to_version
                        and item.owning_package == existing_def.owning_package
                ):
                    raise ConflictingChange(f'The maintscript defines the action {item.command} for'
                                            f' "{obsolete_conffile}" in {mainscript_file}, but there is another'
                                            f' conffile management definition for same path defined already (in the'
                                            f' existing manifest or an migration e.g., inside {mainscript_file})')
                already_present += 1
                continue

            package_definition.add_conffile_management(item)
            manifest_changes += 1

    yield FeatureMigration('dh_installdeb files',
                           migration_changes_manifest=True,
                           warnings=warnings,
                           remove_paths_on_success=paths_to_remove,
                           successful_changes=manifest_changes,
                           already_present=already_present,
                           )


def migrate_tmpfile(manifest: HighLevelManifest,
                    acceptable_migration_issues: AcceptableMigrationIssues,
                    ) -> Iterable[FeatureMigration]:
    warnings = []
    for dctrl_bin in manifest.all_packages:
        dh_config_file = dhe_pkgfile(dctrl_bin, 'tmpfile')
        if dh_config_file is not None:
            # TODO: Support migrating the file to its new location.
            if 'legacy-dh_installtmpfiles-config-name' not in acceptable_migration_issues:
                raise UnsupportedFeature(f'The dh_installtmpfiles file {dh_config_file} is using the "legacy" name. '
                                         ' Please use the ".tmpfiles" (note the "s" at the end) name instead.  This is'
                                         ' also supported by debhelper assuming you are using dh_installtmpfiles'
                                         ' (default in compat 13 or newer).')
            warnings.append(f'TODO: MANUAL MIGRATION of legacy dh_installtmpfiles config "{dh_config_file}" is'
                            ' required.')
    yield FeatureMigration('dh_installtmpfiles config files',
                           migration_changes_manifest=False,
                           warnings=warnings,
                           remove_paths_on_success=tuple(),
                           successful_changes=0,
                           already_present=0,
                           )


def migrate_lintian_overrides_files(manifest: HighLevelManifest,
                                    acceptable_migration_issues: AcceptableMigrationIssues,
                                    ) -> Iterable[FeatureMigration]:
    warnings = []
    for dctrl_bin in manifest.all_packages:
        _dh_config_file(dctrl_bin,
                        'lintian-overrides',
                        'dh_lintian',
                        acceptable_migration_issues,
                        warnings,
                        None,
                        support_executable_files=False,
                        )
    yield FeatureMigration('dh_lintian config files',
                           migration_changes_manifest=False,
                           warnings=warnings,
                           remove_paths_on_success=tuple(),
                           successful_changes=0,
                           already_present=0,
                           )


def migrate_links_files(manifest: HighLevelManifest,
                        acceptable_migration_issues: AcceptableMigrationIssues,
                        ) -> Iterable[FeatureMigration]:
    mutable_manifest = manifest.mutable_manifest
    paths_to_remove = []
    warnings = []
    manifest_changes = 0
    already_present = 0
    for dctrl_bin in manifest.all_packages:
        links_file, content = _dh_config_file(dctrl_bin,
                                              'links',
                                              'dh_link',
                                              acceptable_migration_issues,
                                              warnings,
                                              paths_to_remove,
                                              support_executable_files=True,
                                              )

        if links_file is None:
            continue
        assert content is not None

        package_definition = mutable_manifest.package(dctrl_bin.name)
        defined_symlink = {symlink.symlink_path: symlink.symlink_target for symlink in package_definition.symlinks()}

        seen_symlinks = set()

        for dhe_line in content:
            if len(dhe_line.tokens) != 2:
                raise UnsupportedFeature(f'The dh_link file {links_file} did not have exactly two paths on line'
                                         f' {dhe_line.line_no} (line: "{dhe_line.original_line}"')
            target, source = dhe_line.tokens
            if source in seen_symlinks:
                # According to #934499, this has happened in the wild already
                raise ConflictingChange(f'The {links_file} file defines the link path {source} twice! Please ensure'
                                        ' that it is defined at most once in that file')
            seen_symlinks.add(source)
            # Symlinks in .links are always considered absolute, but you were not required to have a leading slash.
            # However, in the debputy manifest, you can have relative links, so we should ensure it is explicitly
            # absolute.
            if not target.startswith('/'):
                target = '/' + target
            existing_target = defined_symlink.get(source)
            if existing_target is not None:
                if existing_target != target:
                    raise ConflictingChange(f'The symlink "{source}" points to "{target}" in {links_file}, but there is'
                                            f' another symlink with same path pointing to "{existing_target}" defined'
                                            ' already (in the existing manifest or an migration e.g., inside'
                                            f' {links_file})')
                already_present += 1
                continue
            condition = _extract_condition_from_dh_config_line(dhe_line,
                                                               acceptable_migration_issues,
                                                               warnings,
                                                               )
            package_definition.add_symlink(MutableYAMLSymlink.new_symlink(source,
                                                                          target,
                                                                          condition,
                                                                          ))
            manifest_changes += 1
    yield FeatureMigration('dh_link files',
                           migration_changes_manifest=True,
                           warnings=warnings,
                           remove_paths_on_success=paths_to_remove,
                           successful_changes=manifest_changes,
                           already_present=already_present,
                           )


def migrate_dh_hook_targets(_: HighLevelManifest,
                            acceptable_migration_issues: AcceptableMigrationIssues,
                            ) -> Iterable[FeatureMigration]:

    warnings = []
    detected_hook_targets = json.loads(subprocess.check_output(['dh_assistant', 'detect-hook-targets']).decode('utf-8'))
    sample_hook_target = None

    for hook_target_def in detected_hook_targets['hook-targets']:
        if hook_target_def['is-empty']:
            continue
        command = hook_target_def['command']
        if command not in DH_COMMANDS_REPLACED:
            continue
        hook_target = hook_target_def['target-name']
        if sample_hook_target is None:
            sample_hook_target = hook_target
        warnings.append(f'TODO: MANUAL MIGRATION required for hook target {hook_target}')
    if warnings and 'dh-hook-targets' not in acceptable_migration_issues:
        assert sample_hook_target
        raise UnsupportedFeature(f'The debian/rules file contains one or more non empty dh hook targets that will not'
                                 f' be run with the debputy (or zz-debputy) dh sequence. Once of these would be'
                                 f' {sample_hook_target}.', ['dh-hook-targets'])
    yield FeatureMigration('dh hook targets',
                           migration_changes_manifest=False,
                           warnings=warnings,
                           successful_changes=0,
                           already_present=0,
                           )


def migrate_shlibs_file(manifest: HighLevelManifest,
                        _: AcceptableMigrationIssues,
                        ) -> Iterable[FeatureMigration]:
    yield _migrate_packageless_file(manifest, 'shlibs')


def migrate_symbols_file(manifest: HighLevelManifest,
                         _: AcceptableMigrationIssues,
                         ) -> Iterable[FeatureMigration]:
    yield _migrate_packageless_file(manifest, 'symbols')


def _migrate_packageless_file(manifest: HighLevelManifest, basename: str) -> FeatureMigration:
    paths_to_remove = []
    paths_to_rename = []
    warnings = []
    pkgless_variant = f'debian/{basename}'
    main_package = [p for p in manifest.all_packages if p.is_main_package][0]
    main_variant = f'debian/{main_package.name}.{basename}'
    if os.path.isfile(pkgless_variant):
        if os.path.isfile(main_variant):
            if subprocess.call(['cmp', '-s', pkgless_variant, main_variant]) != 0:
                warnings.append(f'TODO: MANUAL MIGRATION - there is a {pkgless_variant} and {main_variant}'
                                f' file. Normally these files are for the same package and there would only be one of'
                                f' them. In this case, they both exist but their content differs. Be advised that'
                                f' debputy tool will use the {main_variant}.')
            else:
                paths_to_remove.append(pkgless_variant)
        else:
            paths_to_rename.append(
                (pkgless_variant, main_variant)
            )
    return FeatureMigration(
        f'Migrate {pkgless_variant} to {main_variant}',
        migration_changes_manifest=False,
        successful_changes=bool(len(paths_to_remove) + len(paths_to_rename)),
        warnings=warnings,
        already_present=bool(len(warnings)),
        rename_paths_on_success=paths_to_rename,
        remove_paths_on_success=paths_to_remove,
    )
