import os.path
import shutil
from typing import Optional, Tuple, Set, Dict

from debputy.debhelper_emulation import dhe_generated_file, dhe_pkgfile, dhe_install_path, debhelper_script_subst, \
    DHConfigFileLine
from debputy.maintscript_snippet import STD_CONTROL_SCRIPTS, MaintscriptSnippetContainer
from debputy.packages import BinaryPackage
from debputy.util import PKGNAME_REGEX, PKGVERSION_REGEX, debputy_autoscript_tagline


def _append_generated_triggers(dest_file: str,
                               generated_triggers: Set[str],
                               ) -> None:
    if not generated_triggers:
        return
    with open(dest_file, 'at', encoding='utf-8') as fd:
        fd.write(f'# Triggers added by {debputy_autoscript_tagline()}\n')
        fd.writelines(f'{t}\n' for t in generated_triggers)


def install_triggers(binary_package: BinaryPackage,
                     root_dir: str,
                     generated_triggers: Set[str]) -> None:
    dest_file = os.path.join(root_dir, 'DEBIAN', 'triggers')

    provided_triggers = dhe_pkgfile(binary_package, 'triggers')
    dh_generated_triggers_file = dhe_generated_file(binary_package, 'triggers', mkdirs=False)

    all_trigger_files = [
        f for f in (provided_triggers, dh_generated_triggers_file) if f is not None and os.path.isfile(f)
    ]

    if not all_trigger_files:
        if generated_triggers:
            _append_generated_triggers(dest_file, generated_triggers)
            os.chmod(dest_file, 0o644)
        return
    if len(all_trigger_files) == 1:
        dhe_install_path(all_trigger_files[0], dest_file, mode=0o0644)
        _append_generated_triggers(dest_file, generated_triggers)
        return
    with open(dest_file, 'wt', encoding='utf-8') as wfd:
        for trigger_file in all_trigger_files:
            with open(trigger_file, 'rt') as rfd:
                shutil.copyfileobj(rfd, wfd)
    _append_generated_triggers(dest_file, generated_triggers)
    os.chmod(dest_file, 0o644)


def _validate_rm_mv_conffile(package: str,
                             config_line: DHConfigFileLine,
                             ) -> Tuple[str, str, Optional[str], Optional[str], Optional[str]]:
    cmd, *args = config_line.tokens
    if '--' in config_line.tokens:
        raise ValueError(f'The maintscripts file "{config_line.config_file}" for {package} includes a "--" in line'
                         f' {config_line.line_no}. The offending line is: {config_line.original_line}')
    if cmd == 'rm_conffile':
        min_args = 1
        max_args = 3
    else:
        min_args = 2
        max_args = 4
    if len(args) > max_args or len(args) < min_args:
        raise ValueError(f'The "{cmd}" command takes at least {min_args} and at most {max_args} arguments.  However,'
                         f' in "{config_line.config_file}" line {config_line.line_no} (for {package}), there'
                         f' are {len(args)} arguments. The offending line is: {config_line.original_line}')

    obsolete_conffile = args[0]
    new_conffile = args[1] if cmd == 'mv_conffile' else None
    prior_version = args[min_args] if len(args) > min_args else None
    owning_package = args[min_args + 1] if len(args) > min_args + 1 else None
    if not obsolete_conffile.startswith('/'):
        raise ValueError(f'The (old-)conffile parameter for {cmd} must be absolute (i.e., start with "/").  However,'
                         f' in "{config_line.config_file}" line {config_line.line_no} (for {package}), it was specified'
                         f' as "{obsolete_conffile}". The offending line is: {config_line.original_line}')
    if new_conffile is not None and not new_conffile.startswith('/'):
        raise ValueError(f'The new-conffile parameter for {cmd} must be absolute (i.e., start with "/").  However,'
                         f' in "{config_line.config_file}" line {config_line.line_no} (for {package}), it was specified'
                         f' as "{new_conffile}". The offending line is: {config_line.original_line}')
    if prior_version is not None and not PKGVERSION_REGEX.fullmatch(prior_version):
        raise ValueError(f'The prior-version parameter for {cmd} must be a valid package version (i.e., match'
                         f' {PKGVERSION_REGEX}).  However, in "{config_line.config_file}" line {config_line.line_no}'
                         f' (for {package}), it was specified as "{prior_version}". The offending line is:'
                         f' {config_line.original_line}')
    if owning_package is not None and not PKGNAME_REGEX.fullmatch(owning_package):
        raise ValueError(f'The package parameter for {cmd} must be a valid package name (i.e., match {PKGNAME_REGEX}).'
                         f'  However, in "{config_line.config_file}" line {config_line.line_no} (for {package}), it'
                         f' was specified as "{owning_package}". The offending line is: {config_line.original_line}')
    return cmd, obsolete_conffile, new_conffile, prior_version, owning_package


def dh_installdeb_emulation(binary_package: BinaryPackage,
                            root_dir: str,
                            generated_triggers: Set[str],
                            maintscript_snippets: Dict[str, MaintscriptSnippetContainer],
                            ) -> None:

    for script in STD_CONTROL_SCRIPTS:
        debhelper_script_subst(binary_package, root_dir, script, maintscript_snippets, None)

    install_triggers(binary_package, root_dir, generated_triggers)
