import dataclasses
from typing import Mapping, Iterable, Dict, List, Optional, Tuple, Callable

from debputy.manifest_path_rules import ManifestPathRule
from debputy.packages import BinaryPackage
from debputy.path_matcher import ExactFileSystemPath
from debputy.plugin.api.impl import PackagerProvidedFileClassSpec
from debputy.substitution import Substitution
from debputy.typing import DirEntryLike


@dataclasses.dataclass(frozen=True, slots=True)
class PackagerProvidedFile:
    source_basename: str
    package_name: str
    installed_as_basename: str
    provided_key: str
    definition: PackagerProvidedFileClassSpec
    match_priority: int = 0

    def compute_dest_name(self) -> str:
        return self.definition.compute_dest_basename(self.installed_as_basename)


def _find_package_name_prefix(binary_packages: Mapping[str, BinaryPackage],
                              main_binary_package: str,
                              max_periods_in_package_name: int,
                              orig_basename: str,
                              ) -> Iterable[Tuple[str, str, bool]]:
    if max_periods_in_package_name < 1:
        prefix, remaining = orig_basename.split('.', 1)
        if prefix in binary_packages:
            yield prefix, remaining, True
        else:
            yield main_binary_package, orig_basename, False
        return

    parts = orig_basename.split('.', max_periods_in_package_name + 1)
    for p in range(len(parts) - 1, 0, -1):
        name = '.'.join(parts[0:p])
        if name in binary_packages:
            remaining = '.'.join(parts[p:])
            yield name, remaining, True
    # main package case
    yield main_binary_package, orig_basename, False


def _split_path(packager_provided_files: Mapping[str, PackagerProvidedFileClassSpec],
                binary_packages: Mapping[str, BinaryPackage],
                main_binary_package: str,
                max_periods_in_package_name: int,
                orig_basename: str,
                ) -> Optional[PackagerProvidedFile]:
    owning_package_name = main_binary_package
    basename = orig_basename
    match_priority = 0
    if '.' not in basename:
        definition = packager_provided_files.get(basename)
        if definition is None:
            return None
        return PackagerProvidedFile(
            source_basename=orig_basename,
            package_name=owning_package_name,
            installed_as_basename=owning_package_name,
            provided_key='.UNNAMED.',
            definition=definition,
            match_priority=match_priority,
        )

    for owning_package_name, basename, explicit_package in _find_package_name_prefix(binary_packages,
                                                                                     main_binary_package,
                                                                                     max_periods_in_package_name,
                                                                                     orig_basename,
                                                                                     ):
        owning_package = binary_packages[owning_package_name]
        match_priority = 1 if explicit_package else 0
        if not owning_package.is_arch_all and '.' in basename:
            remaining, last_word = basename.rsplit('.', 1)
            if last_word == owning_package.resolved_architecture:
                match_priority = 3
                basename = remaining
            elif last_word == owning_package.package_deb_architecture_variable('ARCH_OS'):
                match_priority = 2
                basename = remaining

        definition = packager_provided_files.get(basename)
        if definition is not None:
            return PackagerProvidedFile(
                source_basename=orig_basename,
                package_name=owning_package_name,
                installed_as_basename=owning_package_name,
                provided_key='.UNNAMED.',
                definition=definition,
                match_priority=match_priority,
            )

        install_as_name = basename
        file_class = ''
        while '.' in install_as_name:
            install_as_name, file_class_part = install_as_name.rsplit('.', 1)
            file_class = file_class_part + '.' + file_class if file_class != '' else file_class_part
            definition = packager_provided_files.get(file_class)
            if definition is not None:
                return PackagerProvidedFile(
                    source_basename=orig_basename,
                    package_name=owning_package_name,
                    installed_as_basename=install_as_name,
                    provided_key=install_as_name,
                    definition=definition,
                    match_priority=match_priority,
                )
    return None


def detect_all_packager_provided_files(packager_provided_files: Mapping[str, PackagerProvidedFileClassSpec],
                                       debian_dir: Iterable[DirEntryLike],
                                       binary_packages: Mapping[str, BinaryPackage],
                                       substitution_for: Callable[[str], Substitution],
                                       ) -> Dict[str, List['ManifestPathRule']]:
    main_binary_package = [p.name for p in binary_packages.values() if p.is_main_package][0]
    provided_files: Dict[str, Dict[str, PackagerProvidedFile]] = {n: {} for n in binary_packages}
    max_periods_in_package_name = max(name.count('.') for name in binary_packages)

    for entry in debian_dir:
        if not entry.is_file(follow_symlinks=True):
            continue
        packager_provided_file = _split_path(packager_provided_files,
                                             binary_packages,
                                             main_binary_package,
                                             max_periods_in_package_name,
                                             entry.name)
        if packager_provided_file is None:
            continue
        provided_files_for_package = provided_files[packager_provided_file.package_name]
        existing = provided_files_for_package.get(packager_provided_file.provided_key)
        if existing is not None and existing.match_priority > packager_provided_file.match_priority:
            continue
        provided_files_for_package[packager_provided_file.provided_key] = packager_provided_file

    result = {
        k: [
            ManifestPathRule.packager_provided_file(
                definition_source=f'debian/{x.source_basename} (auto-detected by plugin {x.definition.debputy_plugin_metadata.plugin_name})',
                match_rule=ExactFileSystemPath(
                    f'{substitution_for(k).substitute(x.definition.install_into_directory, f"debian/{x.source_basename}")}/{x.compute_dest_name()}'
                ),
                source_path=f'debian/{x.source_basename}',
                mode=x.definition.default_mode,
            ) for x in v.values()
        ] for k, v in provided_files.items()
    }

    return result
