import argparse
import glob
import os
import re
import sys
import time
from pathlib import Path
from typing import NoReturn, TYPE_CHECKING, Union, Set, FrozenSet, Optional, TypeVar, IO, cast

from debian.deb822 import Deb822

if TYPE_CHECKING:
    from debputy.filesystem_scan import FSPath


T = TypeVar('T')


SLASH_PRUNE = re.compile('//+')
PKGNAME_REGEX = re.compile(r'[a-z0-9][-+.a-z0-9]+', re.ASCII)
PKGVERSION_REGEX = re.compile(r'''
                 (?: \d+ : )?                # Optional epoch
                 \d[0-9A-Za-z.+:~]*          # Upstream version (with no hyphens)
                 (?: - [0-9A-Za-z.+:~]+ )*   # Optional debian revision (+ upstreams versions with hyphens)
''', re.VERBOSE | re.ASCII)
DEFAULT_PACKAGE_TYPE = 'deb'
DBGSYM_PACKAGE_TYPE = 'deb'
UDEB_PACKAGE_TYPE = 'udeb'

POSTINST_DEFAULT_CONDITION = (
    '[ "$1" = "configure" ]'
    ' || [ "$1" = "abort-upgrade" ]'
    ' || [ "$1" = "abort-deconfigure" ]'
    ' || [ "$1" = "abort-remove" ]'
)


_SPACE_RE = re.compile(r'\s')
_DOUBLE_ESCAPEES = re.compile(r'([\n`$"\\])')
_REGULAR_ESCAPEES = re.compile(r'([\s!"$()*+#;<>?@\[\]\\`|~])')
_PROFILE_GROUP_SPLIT = re.compile(r'>\s+<')


def debputy_autoscript_tagline() -> str:
    # TODO: Handle versioning at some point.
    return 'debputy/<unversioned>'


def assume_not_none(x: Optional[T]) -> T:
    if x is None:  # pragma: no cover
        raise ValueError('Internal error: None was given, but the receiver assumed "not None" here')
    return x


_COLOR_CHECKED = None


class _ANSIColor:

    _CODES_AND_STYLES = {
        'bold': '1',
        'red': '31',
        'yellow': '33',
    }

    def __init__(self) -> None:
        self._checked_color = None

    def _color_active(self, fd: IO) -> bool:
        if self._checked_color is None:
            dpkg_or_default = os.environ.get('DPKG_COLORS', 'never' if 'NO_COLOR' in os.environ else 'auto')
            self._checked_color = os.environ.get('DEBPUTY_COLORS', dpkg_or_default)
        if self._checked_color not in {'auto', 'always', 'never'}:
            # Default to auto before invoking _warn to have a sense default apply for that call
            self._checked_color = 'auto'
            _warn(f'Invalid color request for "{self._checked_color}" in either DEBPUTY_COLORS or DPKG_COLORS.'
                  ' Resetting to "auto".')
        return self._checked_color == 'always' or (self._checked_color == 'auto' and fd.isatty())

    def color_phrase(self, fd: IO, color_request: str, phrase: str) -> str:
        if not self._color_active(fd):
            return phrase
        esc_sequence = '\033['
        color_code = ';'.join(self._CODES_AND_STYLES[c] for c in color_request.split())
        if not color_code:
            return phrase
        reset_code = f'{esc_sequence}0m'
        return f'{esc_sequence}{color_code}m{phrase}{reset_code}'


_ANSI_COLOR = _ANSIColor()
del _ANSIColor


def _error(msg: str, *, prog=None) -> 'NoReturn':
    me = os.path.basename(sys.argv[0]) if prog is None else prog
    print(f'{me}: {_ANSI_COLOR.color_phrase(sys.stderr, "bold red", "error")}: {msg}', file=sys.stderr)
    sys.exit(1)


def _warn(msg: str, *, prog=None) -> None:
    me = os.path.basename(sys.argv[0]) if prog is None else prog
    print(f'{me}: {_ANSI_COLOR.color_phrase(sys.stderr, "bold yellow", "warning")}: {msg}', file=sys.stderr)


class ColorizedArgumentParser(argparse.ArgumentParser):

    def error(self, message: str) -> NoReturn:
        self.print_usage(sys.stderr)
        _error(message, prog=self.prog)


def ensure_dir(path: str) -> None:
    if not os.path.isdir(path):
        os.makedirs(path, mode=0o755, exist_ok=True)


def _clean_path(orig_p: str) -> str:
    p = SLASH_PRUNE.sub('/', orig_p)
    if '.' in p:
        path_base = p
        # We permit a single leading "./" because we add that when we normalize a path, and we want normalization
        # of a normalized path to be a no-op.
        if path_base.startswith("./"):
            path_base = path_base[2:]
            assert path_base
        for segment in path_base.split('/'):
            if segment in ('.', '..'):
                raise ValueError('Please provide paths that are normalized (i.e., no ".." or ".").'
                                 f' Offending input "{orig_p}"')
    return p


def _normalize_path(path: str, with_prefix=True) -> str:
    path = path.strip('/')
    if not path or path == '.':
        return '.'
    if '//' in path or '.' in path:
        path = _clean_path(path)
    if with_prefix ^ path.startswith('./'):
        if with_prefix:
            path = './' + path
        else:
            path = path[2:]
    return path


def _normalize_link_target(link_target: str) -> str:
    link_target = SLASH_PRUNE.sub('/', link_target.lstrip('/'))
    result = []
    for segment in link_target.split('/'):
        if segment in ('.', ''):
            # Ignore these - the empty string is generally a trailing slash
            continue
        if segment == '..':
            # We ignore "root escape attempts" like the OS would (mapping /.. -> /)
            if result:
                result.pop()
        else:
            result.append(segment)
    return '/'.join(result)


def _backslash_escape(m: re.Match[str]) -> str:
    return '\\' + m.group(0)


def _escape_shell_word(w: str) -> str:
    if _SPACE_RE.match(w):
        _DOUBLE_ESCAPEES.sub(_backslash_escape, w)
        return f'"{w}"'
    return _REGULAR_ESCAPEES.sub(_backslash_escape, w)


def escape_shell(*args: str) -> str:
    return ' '.join(_escape_shell_word(w) for w in args)


def debian_policy_normalize_symlink_target(link_path: str,
                                           link_target: str,
                                           normalize_link_path=False,
                                           ) -> str:
    if normalize_link_path:
        link_path = _normalize_path(link_path)
    elif not link_path.startswith('./'):
        raise ValueError("Link part was not normalized")

    link_path = link_path[2:]

    if not link_target.startswith('/'):
        link_target = '/' + os.path.dirname(link_path) + '/' + link_target

    link_path_parts = link_path.split('/')
    link_target_parts = [s for s in _normalize_link_target(link_target).split('/') if s != '.']

    assert link_path_parts

    if link_target_parts and link_path_parts[0] == link_target_parts[0]:
        # Per Debian Policy, must be relative

        # First determine the length of the overlap
        common_segment_count = 1
        shortest_path_length = min(len(link_target_parts), len(link_path_parts))
        while (common_segment_count < shortest_path_length
                and link_target_parts[common_segment_count] == link_path_parts[common_segment_count]):
            common_segment_count += 1

        if common_segment_count == shortest_path_length and len(link_path_parts) - 1 == len(link_target_parts):
            normalized_link_target = '.'
        else:
            up_dir_count = len(link_path_parts) - 1 - common_segment_count
            normalized_link_target_parts = []
            if up_dir_count:
                up_dir_part = '../' * up_dir_count
                # We overshoot with a single '/', so rstrip it away
                normalized_link_target_parts.append(up_dir_part.rstrip('/'))
            # Add the relevant down parts
            normalized_link_target_parts.extend(link_target_parts[common_segment_count:])

            normalized_link_target = '/'.join(normalized_link_target_parts)
    else:
        # Per Debian Policy, must be absolute
        normalized_link_target = '/' + '/'.join(link_target_parts)

    return normalized_link_target


def apply_symbolic_mode(tar_path: 'FSPath', base_mode: int, symbolic_mode: str) -> int:
    final_mode = base_mode
    special_bit_marker = 0o100_000
    sticky_bit = 0o01000
    setuid_bit = 0o04000
    setgid_bit = 0o02000
    mode_group_flag = 0o7
    bits = {
        'r': 0o4,
        'w': 0o2,
        'x': 0o1,
        'X': 0o1 if tar_path.is_dir or base_mode & 0o111 else 0o0,
        's': special_bit_marker,
        't': sticky_bit,
    }
    modifiers = {
        '+',
        '-',
        '=',
    }
    for orig_part in symbolic_mode.split(','):
        part_mode = 0
        part = orig_part
        subjects = set()
        while part and part[0] in ('u', 'g', 'o', 'a'):
            subject = part[0]
            if subject == 'a':
                subjects = {'u', 'g', 'o'}
            else:
                subjects.add(subject)
            part = part[1:]
        if not subjects:
            subjects = {'u', 'g', 'o'}

        if part and part[0] in modifiers:
            modifier = part[0]
        elif not part:
            raise ValueError(f'Invalid symbolic mode - expected [+-=] to be present (from "{orig_part}")')
        else:
            raise ValueError(f'Invalid symbolic mode - expected "{part[0]}" to be one of [+-=]'
                             f' (from "{orig_part}")')
        part = part[1:]
        if part in ('u', 'g', 'o'):
            raise NotImplementedError("Sorry, we do not support referencing an existing subject's permissions (a=u)"
                                      " in symbolic modes")
        while part and part[0] in bits:
            part_mode |= bits[part[0]]
            part = part[1:]

        if part:
            raise ValueError(f'Invalid symbolic mode - Could not parse "{part[0]}" from "{orig_part}"')

        mod_mode = 0
        if 'u' in subjects:
            mod_mode |= (part_mode & mode_group_flag) << 6
        elif modifier == '=':
            mod_mode |= final_mode & (mode_group_flag << 6) | (final_mode & setuid_bit)
        if 'g' in subjects:
            mod_mode |= (part_mode & mode_group_flag) << 3
        elif modifier == '=':
            mod_mode |= final_mode & (mode_group_flag << 3) | (final_mode & setgid_bit)
        if 'o' in subjects:
            mod_mode |= (part_mode & mode_group_flag)
        elif modifier == '=':
            mod_mode |= final_mode & mode_group_flag
        if part_mode & special_bit_marker:
            if 'u' in subjects:
                mod_mode |= setuid_bit
            if 'g' in subjects:
                mod_mode |= setgid_bit
        if part_mode & sticky_bit:
            mod_mode |= sticky_bit

        if modifier == '+':
            final_mode |= mod_mode
        elif modifier == '-':
            final_mode &= ~mod_mode
        elif modifier == '=':
            # FIXME: Handle "unmentioned directory's setgid/setuid bits"
            final_mode = mod_mode
    return final_mode


def glob_escape(replacement_value: str) -> str:
    if not glob.has_magic(replacement_value) or '{' not in replacement_value:
        return replacement_value
    return replacement_value.replace('[', '[[]') \
        .replace(']', '[]]') \
        .replace('*', '[*]') \
        .replace('?', '[?]') \
        .replace('{', '[{]') \
        .replace('}', '[}]')


# TODO: This logic should probably be moved to `python-debian`
def active_profiles_match(profiles_raw: str,
                          active_build_profiles: Union[Set[str], FrozenSet[str]],
                          ) -> bool:
    profiles_raw = profiles_raw.strip()
    if profiles_raw[0] != '<' or profiles_raw[-1] != '>' or profiles_raw == '<>':
        raise ValueError('Invalid Build-Profiles: Must start start and end with "<" + ">" but cannot be a literal "<>"')
    profile_groups = _PROFILE_GROUP_SPLIT.split(profiles_raw[1:-1])
    for profile_group_raw in profile_groups:
        should_process_package = True
        for profile_name in profile_group_raw.split():
            negation = False
            if profile_name[0] == '!':
                negation = True
                profile_name = profile_name[1:]

            matched_profile = profile_name in active_build_profiles
            if matched_profile == negation:
                should_process_package = False
                break

        if should_process_package:
            return True

    return False


def resolve_source_date_epoch(command_line_value: Optional[int]) -> int:
    mtime = command_line_value
    if mtime is None and 'SOURCE_DATE_EPOCH' in os.environ:
        sde_raw = os.environ['SOURCE_DATE_EPOCH']
        if sde_raw == '':
            _error("SOURCE_DATE_EPOCH is set but empty.")
        mtime = int(sde_raw)
    if mtime is None:
        mtime = int(time.time())
    return mtime


def compute_output_filename(package_root_dir: str, is_udeb: bool) -> str:
    with open(os.path.join(package_root_dir, 'DEBIAN', 'control'), 'rt') as fd:
        control_file = Deb822(fd)

    package_name = control_file['Package']
    package_version = control_file['Version']
    package_architecture = control_file['Architecture']
    extension = control_file.get('Package-Type') or 'deb'
    if ':' in package_version:
        package_version = package_version.split(':', 1)[1]
    if is_udeb:
        extension = 'udeb'

    return f'{package_name}_{package_version}_{package_architecture}.{extension}'


_SCRATCH_DIR = None


def scratch_dir() -> str:
    global _SCRATCH_DIR
    if _SCRATCH_DIR is not None:
        return _SCRATCH_DIR
    debputy_scratch_dir = 'debian/.debputy/scratch-dir'
    is_debputy_dir = True
    if os.path.isdir('debian/.debputy'):
        _SCRATCH_DIR = debputy_scratch_dir
    elif os.path.isdir('debian/.debhelper'):
        _SCRATCH_DIR = 'debian/.debhelper/_debputy/scratch_dir'
        is_debputy_dir = False
    else:
        _SCRATCH_DIR = debputy_scratch_dir
    ensure_dir(_SCRATCH_DIR)
    if is_debputy_dir:
        Path('debian/.debputy/.gitignore').write_text('*\n')
    return _SCRATCH_DIR
